/**
 * File:
 *  main.c
 *
 * Notes:
 *  This application demonstrates the modbus slave implementation
 *  Uses UART1 routines for communicating with serial RS232 port
 *
 * History:
 *  10/06/2012 Written by Simon Platten
 *  30/06/2012 Added mapping parameters to registers to allow 8 bit data ports
 *             to be mapped into 16 bit registers, also changed the number
 *             of holding registers to 2 and the number of input registers to 3.
 *  02/07/2012 Re-wrote I/O structures and functions, all I/O is now added
 *             using addModbusBlock.
 *  03/07/2012 Implementec callbacks and changed I/O type array sizes to:
 *                     coils                8
 *                     status bits          8
 *                     holding registers    1
 *                     input registers      1
 * 06/07/2012 Implemented analogue input via POT
 */
#include "modbus.h"

static volatile modbusBlockDef coilsBlock,
                               statusBitsBlock, 
                               inputRegsBlock, 
                               holdingRegsBlock;
static volatile uint aryuintInputRegs[1], aryuintHoldingRegs[1];
static volatile byte arybytCoils[1], arybytStatusBits[1];
/**
 * Interrupt Service Routine
 */
void interrupt() {
  decodePacket();
}
/**
 * The coils call-back, called when a coil is modified either by function
 * 5 or 15
 */
void coilsCallback(modbusBlockDef* pBlock) {
// In this demo there are 20 coils, we will transfer the state of the coils
// to PORT B which has been set-up for output.
  LATB = ((byte*)pBlock->paryData)[0];
}
void holdingRegCallback(modbusBlockDef* pBlock) {
// In this demo there are 1 holding register, we will transfer the state
// to PORT A which has been set-up for input
  LATA = ((uint*)pBlock->paryData)[0];
}
/**
 * Entry point
 */
void main() {
#ifdef P18F45K22
  ANSELA = 0x02;             // Configure RA1 pin as analog
  ANSELB = 0;
  ANSELC = 0;
  ANSELE = 0;
#endif
// Configure PORTA, PORTB, PORTC and PORTE as output
  TRISA = 0x02;                  // Set RA1 pin as input
  TRISB = 0;
  TRISC = 0;
  TRISE = 0;
// Configure PORTD as input
  TRISD = 0xff;
// Initialise ports
  LATB = 0;
  LATC = 0;
  LATE = 0;
// Initialise the serial port
  modbusSerialInit(BAUD_115200, 1, 1);
// Make sure the data areas are all cleared
  memset(arybytCoils,        0, sizeof(arybytCoils));
  memset(arybytStatusBits,   0, sizeof(arybytStatusBits));
  memset(aryuintHoldingRegs, 0, sizeof(aryuintHoldingRegs));
  memset(aryuintInputRegs,   0, sizeof(aryuintInputRegs));
// Create the various data I/O blocks
  addModbusBlock(1, COILS,             &coilsBlock,       1, 8,
                 (void*)arybytCoils, coilsCallback);
  addModbusBlock(1, STATUS_INPUTS,     &statusBitsBlock,  1, 8,
                 (void*)arybytStatusBits, NULL);
  addModbusBlock(1, HOLDING_REGISTERS, &holdingRegsBlock, 1, 1,
                 (void*)aryuintHoldingRegs, holdingRegCallback);
  addModbusBlock(1, INPUT_REGISTERS,   &inputRegsBlock,   1, 1,
                 (void*)aryuintInputRegs, NULL);
// Endless loop
  while (1) {
// Copy the ports to the input status bits
    arybytStatusBits[0] = PORTB;
// Read the ADC
    aryuintInputRegs[0] = ADC_Read(1);  // get ADC value from 1st channel
// Any updates?
    serviceIOBlocks();
  }
}